@@ -40,6 +40,13 @@ SRMBuffer *srmBufferCreateFromDMA(SRMCore *core, SRMDevice *allocator, SRMBuffer
     }
 
     SRMBuffer *buffer = srmBufferCreate(core, allocator);
+
+    if (!buffer->allocator->eglExtensions.EXT_image_dma_buf_import_modifiers)
+    {
+        SRMError("Failed to import DMA buffer EXT_image_dma_buf_import_modifiers extension not available.");
+        goto fail;
+    }
+
     buffer->src = SRM_BUFFER_SRC_DMA;
     buffer->format = dmaData->format;
     buffer->width = dmaData->width;
@@ -57,9 +64,22 @@ SRMBuffer *srmBufferCreateFromDMA(SRMCore *core, SRMDevice *allocator, SRMBuffer
     buffer->target = srmFormatIsInList(buffer->allocator->dmaRenderFormats,
                                         buffer->format, buffer->modifiers[0]) ? GL_TEXTURE_2D : GL_TEXTURE_EXTERNAL_OES;
 
-    //srmBufferGetTextureID(buffer->allocator, buffer);
+    if (!buffer->allocator->glExtensions.OES_EGL_image && buffer->target == GL_TEXTURE_2D)
+    {
+        SRMError("Failed to import DMA buffer with GL_TEXTURE_2D target, OES_EGL_image extension not available.");
+        goto fail;
+    }
+
+    if (!buffer->allocator->glExtensions.OES_EGL_image_external && buffer->target == GL_TEXTURE_EXTERNAL_OES)
+    {
+        SRMError("Failed to import DMA buffer with GL_TEXTURE_EXTERNAL_OES target, OES_EGL_image_external extension not available.");
+        goto fail;
+    }
 
     return buffer;
+fail:
+    free(buffer);
+    return NULL;
 }
 
 SRMBuffer *srmBufferCreateFromCPU(SRMCore *core, SRMDevice *allocator,
@@ -79,26 +99,35 @@ SRMBuffer *srmBufferCreateFromCPU(SRMCore *core, SRMDevice *allocator,
     buffer->height = height;
     buffer->format = format;
     buffer->target = GL_TEXTURE_2D;
+    buffer->modifiers[0] = DRM_FORMAT_MOD_LINEAR;
 
     const SRMGLFormat *glFmt;
 
     UInt8 supportLinear = 0;
 
-    SRMListForeach(fmtIt, buffer->allocator->dmaTextureFormats)
+    /* Seems like Nouveau or Nvidia + GBM = black textures even if no errors are reported.
+     * Fallback to glTexImage2D instead. */
+    if (buffer->allocator->glExtensions.OES_EGL_image &&
+        !core->forceGlesCPUBufferAllocation &&
+        buffer->allocator->driver != SRM_DEVICE_DRIVER_nouveau &&
+        buffer->allocator->driver != SRM_DEVICE_DRIVER_nvidia)
     {
-        SRMFormat *fmt = srmListItemGetData(fmtIt);
-
-        if (fmt->format == format && fmt->modifier == DRM_FORMAT_MOD_LINEAR)
+        SRMListForeach(fmtIt, buffer->allocator->dmaRenderFormats)
         {
-            // Cache format
-            if (fmtIt != srmListGetFront(buffer->allocator->dmaTextureFormats))
-            {
-                srmListRemoveItem(buffer->allocator->dmaTextureFormats, fmtIt);
-                srmListPrependData(buffer->allocator->dmaTextureFormats, fmt);
-            }
+            SRMFormat *fmt = srmListItemGetData(fmtIt);
 
-            supportLinear = 1;
-            break;
+            if (fmt->format == format && fmt->modifier == DRM_FORMAT_MOD_LINEAR)
+            {
+                // Cache most used formats
+                if (fmtIt != srmListGetFront(buffer->allocator->dmaRenderFormats))
+                {
+                    srmListRemoveItem(buffer->allocator->dmaRenderFormats, fmtIt);
+                    srmListPrependData(buffer->allocator->dmaRenderFormats, fmt);
+                }
+
+                supportLinear = 1;
+                break;
+            }
         }
     }
 
@@ -107,19 +136,15 @@ SRMBuffer *srmBufferCreateFromCPU(SRMCore *core, SRMDevice *allocator,
     if (!supportLinear)
         goto glesOnly;
 
-    buffer->modifiers[0] = DRM_FORMAT_MOD_LINEAR;
     buffer->bo = gbm_bo_create_with_modifiers(buffer->allocator->gbm,
                                               width,
                                               height,
                                               format,
                                               &buffer->modifiers[0],
                                               1);
-
     if (!buffer->bo)
     {
         SRMWarning("[SRMBuffer] gbm_bo_create_with_modifiers failed.");
-
-        // Try to use linear so that can be mapped
         buffer->flags = GBM_BO_USE_RENDERING | GBM_BO_USE_LINEAR | GBM_BO_USE_SCANOUT;
         buffer->bo = gbm_bo_create(buffer->allocator->gbm, width, height, format, buffer->flags);
     }
@@ -148,10 +173,6 @@ SRMBuffer *srmBufferCreateFromCPU(SRMCore *core, SRMDevice *allocator,
     buffer->modifiers[0] = gbm_bo_get_modifier(buffer->bo);
     buffer->strides[0] = gbm_bo_get_stride(buffer->bo);
     buffer->pixelSize = buffer->bpp/8;
-
-    buffer->target = srmFormatIsInList(buffer->allocator->dmaRenderFormats,
-                                        buffer->format, buffer->modifiers[0]) ? GL_TEXTURE_2D : GL_TEXTURE_EXTERNAL_OES;
-
     buffer->fds[0] = srmBufferGetDMAFDFromBO(buffer->allocator, buffer->bo);
 
     if (buffer->fds[0] >= 0)
@@ -165,6 +186,7 @@ SRMBuffer *srmBufferCreateFromCPU(SRMCore *core, SRMDevice *allocator,
 
             if (buffer->map == MAP_FAILED)
             {
+                buffer->map = NULL;
                 SRMWarning("[%s] Directly mapping buffer DMA fd failed. Trying gbm_bo_map.", buffer->allocator->name);
                 goto gbmMap;
             }
@@ -191,12 +213,11 @@ SRMBuffer *srmBufferCreateFromCPU(SRMCore *core, SRMDevice *allocator,
 
     if (!buffer->map)
     {
+        buffer->mapData = NULL;
         SRMWarning("[%s] Failed to map DMA FD. Tying gbm_bo_write instead.", buffer->allocator->name);
         goto gbmWrite;
     }
 
-    /* SRMDebug("[%s] Buffer mapped with gbm_bo_map().", buffer->allocator->name); */
-
     mapWrite:
 
     buffer->offsets[0] = gbm_bo_get_offset(buffer->bo, 0);
@@ -205,11 +226,6 @@ SRMBuffer *srmBufferCreateFromCPU(SRMCore *core, SRMDevice *allocator,
 
     pthread_mutex_unlock(&buffer->mutex);
     srmBufferWrite(buffer, stride, 0, 0, width, height, pixels);
-    pthread_mutex_lock(&buffer->mutex);
-
-    /* SRMDebug("[%s] CPU buffer created using mapping.", buffer->allocator->name); */
-
-    pthread_mutex_unlock(&buffer->mutex);
 
     return buffer;
 
@@ -290,7 +306,6 @@ SRMBuffer *srmBufferCreateFromCPU(SRMCore *core, SRMDevice *allocator,
     }
     else
     {
-        UInt8 tmp[width*height*buffer->pixelSize];
         glTexImage2D(GL_TEXTURE_2D,
                      0,
                      glFmt->glInternalFormat,
@@ -299,7 +314,7 @@ SRMBuffer *srmBufferCreateFromCPU(SRMCore *core, SRMDevice *allocator,
                      0,
                      glFmt->glFormat,
                      glFmt->glType,
-                     tmp);
+                     NULL);
     }
 
     SRMDebug("[%s] CPU buffer created using glTexImage2D.", buffer->allocator->name);
@@ -328,6 +343,13 @@ SRMBuffer *srmBufferCreateFromWaylandDRM(SRMCore *core, void *wlBuffer)
     SRMBuffer *buffer = srmBufferCreate(core, NULL);
 
     pthread_mutex_lock(&buffer->mutex);
+
+    if (!buffer->allocator->eglExtensions.KHR_image_pixmap)
+    {
+        SRMError("Can not create buffer from GBM bo. KHR_image_pixmap extension not available.");
+        return NULL;
+    }
+
     buffer->src = SRM_BUFFER_SRC_WL_DRM;
 
     buffer->bo = gbm_bo_import(buffer->allocator->gbm, GBM_BO_IMPORT_WL_BUFFER, wlBuffer, GBM_BO_USE_RENDERING);
@@ -366,9 +388,15 @@ SRMBuffer *srmBufferCreateFromWaylandDRM(SRMCore *core, void *wlBuffer)
 
 GLuint srmBufferGetTextureID(SRMDevice *device, SRMBuffer *buffer)
 {
+    if (!device || !buffer)
+    {
+        SRMError("Invalid parameters passed to srmBufferGetTextureID().");
+        return 0;
+    }
+
     if (buffer->src == SRM_BUFFER_SRC_WL_DRM && device != buffer->allocator)
     {
-        SRMError("[%s] wl_drm buffers can only be accessed from allocator device.");
+        SRMError("[%s] wl_drm buffers can only be accessed from allocator device.", device->name);
         return 0;
     }
 
@@ -400,6 +428,18 @@ GLuint srmBufferGetTextureID(SRMDevice *device, SRMBuffer *buffer)
         }
     }
 
+    if (buffer->target == GL_TEXTURE_2D && !device->glExtensions.OES_EGL_image)
+    {
+        SRMError("[%s] Failed to get texture id from EGL image, OES_EGL_image extension not available.", device->name);
+        return 0;
+    }
+
+    if (buffer->target == GL_TEXTURE_EXTERNAL_OES && !device->glExtensions.OES_EGL_image_external)
+    {
+        SRMError("[%s] Failed to get texture id from EGL image, OES_EGL_image_external extension not available.", device->name);
+        return 0;
+    }
+
     pthread_mutex_lock(&buffer->mutex);
 
     // Creates the texture
@@ -410,7 +450,7 @@ GLuint srmBufferGetTextureID(SRMDevice *device, SRMBuffer *buffer)
     UInt32 index = 0;
     EGLAttrib imageAttribs[sizeof(EGLAttrib)*(8 + 10*buffer->planesCount) + 1];
 
-    if (device == buffer->allocator && buffer->bo)
+    if (device->eglExtensions.KHR_image_pixmap && device == buffer->allocator && buffer->bo && device)
     {
         imageAttribs[0] = EGL_IMAGE_PRESERVED_KHR;
         imageAttribs[1] = EGL_TRUE;
@@ -421,18 +461,6 @@ GLuint srmBufferGetTextureID(SRMDevice *device, SRMBuffer *buffer)
             goto skipDMA;
     }
     
-    if (!buffer->allocator->capPrimeExport)
-    {
-        SRMError("srmBufferGetTextureID failed. Allocator device (%s) has not the PRIME export cap.", buffer->allocator->name);
-        goto skipDMA;
-    }
-
-    if (!device->capPrimeImport)
-    {
-        SRMError("srmBufferGetTextureID failed. Target device (%s) has not the PRIME import cap.", device->name);
-        goto skipDMA;
-    }
-
     if (buffer->fds[0] == -1 && buffer->bo)
     {
         buffer->fds[0] = srmBufferGetDMAFDFromBO(buffer->allocator, buffer->bo);
@@ -441,7 +469,7 @@ GLuint srmBufferGetTextureID(SRMDevice *device, SRMBuffer *buffer)
             buffer->fds[i] = buffer->fds[0];
     }
 
-    if (buffer->fds[0] == -1)
+    if (buffer->fds[0] == -1 || !device->eglExtensions.EXT_image_dma_buf_import_modifiers)
         goto skipDMA;
 
     imageAttribs[index++] = EGL_WIDTH;
@@ -569,19 +597,6 @@ UInt8 srmBufferWrite(SRMBuffer *buffer, UInt32 stride, UInt32 dstX, UInt32 dstY,
         }
         else
         {
-            #if SRM_PAR_CPY == 1
-            if (buffer->core->copyThreadsCount)
-            {
-                srmCoreCopy(buffer->core, (UInt8*)pixels, dst, stride, buffer->strides[0], dstWidth * buffer->pixelSize, dstHeight);
-
-                recheck:
-                for (UInt8 i = 0; i < buffer->core->copyThreadsCount; i++)
-                    if (!buffer->core->copyThreads[i].finished)
-                        goto recheck;
-            }
-            else
-            #endif
-
             for (UInt32 i = 0; i < dstHeight; i++)
             {
                 memcpy(dst,
@@ -653,6 +668,12 @@ SRMBuffer *srmBufferCreateFromGBM(SRMCore *core, struct gbm_bo *bo)
         return NULL;
     }
 
+    if (!allocDev->eglExtensions.KHR_image_pixmap)
+    {
+        SRMError("Can not create buffer from GBM bo. KHR_image_pixmap extension not available.");
+        return NULL;
+    }
+
     SRMBuffer *buffer = srmBufferCreate(core, allocDev);
     buffer->src = SRM_BUFFER_SRC_GBM;
     buffer->planesCount = gbm_bo_get_plane_count(bo);
