@@ -1,4 +1,4 @@
-#include "private/modes/SRMRenderModeCommon.h"
+#include <private/modes/SRMRenderModeCommon.h>
 #include <private/SRMConnectorPrivate.h>
 #include <private/SRMDevicePrivate.h>
 #include <private/SRMEncoderPrivate.h>
@@ -101,10 +101,12 @@ UInt8 srmConnectorSetCursor(SRMConnector *connector, UInt8 *pixels)
         if (connector->cursorVisible == 0)
             return 1;
 
-        if (connector->device->clientCapAtomic)
+        if (connector->device->clientCapAtomic && connector->currentVSync)
         {
+            pthread_mutex_lock(&connector->propsMutex);
             connector->cursorVisible = 0;
-            connector->atomicCursorHasChanges |= SRM_CURSOR_ATOMIC_CHANGE_VISIBILITY;
+            connector->atomicChanges |= SRM_ATOMIC_CHANGE_CURSOR_VISIBILITY;
+            pthread_mutex_unlock(&connector->propsMutex);
             pthread_cond_signal(&connector->repaintCond);
         }
         else
@@ -120,8 +122,10 @@ UInt8 srmConnectorSetCursor(SRMConnector *connector, UInt8 *pixels)
 
     gbm_bo_write(connector->cursorBOPending, pixels, 64*64*4);
 
-    if (connector->device->clientCapAtomic)
-        connector->atomicCursorHasChanges |= SRM_CURSOR_ATOMIC_CHANGE_BUFFER;
+    pthread_mutex_lock(&connector->propsMutex);
+
+    if (connector->device->clientCapAtomic && connector->currentVSync)
+        connector->atomicChanges |= SRM_ATOMIC_CHANGE_CURSOR_BUFFER;
     else
     {
         struct gbm_bo *tmpBo = connector->cursorBO;
@@ -133,15 +137,16 @@ UInt8 srmConnectorSetCursor(SRMConnector *connector, UInt8 *pixels)
         connector->cursorFBPending = tmpFb;
     }
 
-    if (connector->cursorVisible == 1 && connector->device->clientCapAtomic && connector->atomicCursorHasChanges)
+    if (connector->cursorVisible == 1 && connector->device->clientCapAtomic && connector->atomicChanges && connector->currentVSync)
     {
+        pthread_mutex_unlock(&connector->propsMutex);
         pthread_cond_signal(&connector->repaintCond);
         return 1;
     }
 
-    if (connector->device->clientCapAtomic)
+    if (connector->device->clientCapAtomic && connector->currentVSync)
     {
-        connector->atomicCursorHasChanges |= SRM_CURSOR_ATOMIC_CHANGE_VISIBILITY;
+        connector->atomicChanges |= SRM_ATOMIC_CHANGE_CURSOR_VISIBILITY;
         pthread_cond_signal(&connector->repaintCond);
     }
     else
@@ -154,7 +159,7 @@ UInt8 srmConnectorSetCursor(SRMConnector *connector, UInt8 *pixels)
     }
 
     connector->cursorVisible = 1;
-
+    pthread_mutex_unlock(&connector->propsMutex);
     return 1;
 }
 
@@ -166,11 +171,13 @@ UInt8 srmConnectorSetCursorPos(SRMConnector *connector, Int32 x, Int32 y)
     if (connector->cursorX == x && connector->cursorY == y)
         return 1;
 
-    if (connector->device->clientCapAtomic)
+    if (connector->device->clientCapAtomic && connector->currentVSync)
     {
+        pthread_mutex_lock(&connector->propsMutex);
         connector->cursorX = x;
         connector->cursorY = y;
-        connector->atomicCursorHasChanges |= SRM_CURSOR_ATOMIC_CHANGE_POSITION;
+        connector->atomicChanges |= SRM_ATOMIC_CHANGE_CURSOR_POSITION;
+        pthread_mutex_unlock(&connector->propsMutex);
         pthread_cond_signal(&connector->repaintCond);
     }
     else
@@ -312,6 +319,25 @@ UInt8 srmConnectorInitialize(SRMConnector *connector, SRMConnectorInterface *int
     connector->renderInitResult = 0;
     connector->firstPageFlip = 1;
 
+    UInt64 gammaSize = srmCrtcGetGammaSize(connector->currentCrtc);
+
+    if (gammaSize > 0)
+    {
+        SRMDebug("[%s] Connector (%d) gamma size = %d.",
+                connector->device->name,
+                connector->id,
+                (UInt32)gammaSize);
+
+        if (connector->device->clientCapAtomic && connector->currentCrtc->propIDs.GAMMA_LUT_SIZE)
+            connector->gamma = malloc(gammaSize * sizeof(*connector->gamma));
+    }
+    else
+    {
+        SRMDebug("[%s] Connector (%d) does not support gamma correction.",
+                 connector->device->name,
+                 connector->id);
+    }
+
     if (pthread_create(&connector->renderThread, NULL, srmConnectorRenderThread, connector))
     {
         SRMError("Could not start render thread for device %s connector %d.", connector->device->name, connector->id);
@@ -336,23 +362,13 @@ UInt8 srmConnectorInitialize(SRMConnector *connector, SRMConnectorInterface *int
     return 1;
 
 fail:
-    connector->renderThread = 0;
-    connector->currentEncoder = NULL;
-    connector->currentCrtc = NULL;
-    connector->currentPrimaryPlane = NULL;
-    connector->currentCursorPlane = NULL;
-
-    bestEncoder->currentConnector = NULL;
-    bestCrtc->currentConnector = NULL;
-    bestPrimaryPlane->currentConnector = NULL;
+    connector->state = SRM_CONNECTOR_STATE_UNINITIALIZED;
 
     if (bestCursorPlane)
         bestCursorPlane->currentConnector = NULL;
 
-    connector->interfaceData = NULL;
-    connector->interface = NULL;
+    srmConnectorRenderThreadCleanUp(connector);
 
-    connector->state = SRM_CONNECTOR_STATE_UNINITIALIZED;
     return 0;
 }
 
@@ -394,33 +410,7 @@ void srmConnectorUninitialize(SRMConnector *connector)
         usleep(1000);
     }
 
-    if (connector->currentCrtc)
-    {
-        connector->currentCrtc->currentConnector = NULL;
-        connector->currentCrtc = NULL;
-    }
-
-    if (connector->currentEncoder)
-    {
-        connector->currentEncoder->currentConnector = NULL;
-        connector->currentEncoder = NULL;
-    }
-
-    if (connector->currentPrimaryPlane)
-    {
-        connector->currentPrimaryPlane->currentConnector = NULL;
-        connector->currentPrimaryPlane = NULL;
-    }
-
-    if (connector->currentCursorPlane)
-    {
-        connector->currentCursorPlane->currentConnector = NULL;
-        srmConnectorSetCursorPlaneToNeededConnector(connector->currentCursorPlane);
-        connector->currentCursorPlane = NULL;
-    }
-
-    connector->interfaceData = NULL;
-    connector->interface = NULL;
+    srmConnectorRenderThreadCleanUp(connector);
 
     SRMDebug("[%s] Connector (%d) %s, %s, %s uninitialized.",
              connector->device->name,
@@ -461,14 +451,14 @@ UInt8 srmConnectorSuspend(SRMConnector *connector)
         case SRM_CONNECTOR_STATE_INITIALIZED:
         {
             connector->state = SRM_CONNECTOR_STATE_SUSPENDING;
-            connector->atomicCursorHasChanges = 0;
+            connector->atomicChanges = 0;
             pthread_mutex_unlock(&connector->stateMutex);
             return srmConnectorSuspend(connector);
         }
         default:
         {
             srmConnectorUnlockRenderThread(connector, 1);
-            connector->atomicCursorHasChanges = 0;
+            connector->atomicChanges = 0;
             pthread_mutex_unlock(&connector->stateMutex);
             usleep(10000);
             return srmConnectorSuspend(connector);
@@ -536,6 +526,9 @@ SRMBuffer *srmConnectorGetBuffer(SRMConnector *connector, UInt32 bufferIndex)
 
 UInt8 srmConnectorHasBufferDamageSupport(SRMConnector *connector)
 {
+    if (connector->currentPrimaryPlane && connector->currentPrimaryPlane->propIDs.FB_DAMAGE_CLIPS)
+        return 1;
+
     SRM_RENDER_MODE renderMode = srmDeviceGetRenderMode(connector->device);
 
     if (renderMode == SRM_RENDER_MODE_ITSELF)
@@ -547,22 +540,165 @@ UInt8 srmConnectorHasBufferDamageSupport(SRMConnector *connector)
 
 UInt8 srmConnectorSetBufferDamage(SRMConnector *connector, SRMRect *rects, Int32 n)
 {
-    if (!srmConnectorHasBufferDamageSupport(connector))
+    if (!connector->currentPrimaryPlane || !srmConnectorHasBufferDamageSupport(connector))
         return 0;
 
-    if (connector->damageRects)
+    if (connector->damageBoxes)
     {
-        free(connector->damageRects);
-        connector->damageRects = NULL;
-        connector->damageRectsCount = 0;
+        free(connector->damageBoxes);
+        connector->damageBoxes = NULL;
+        connector->damageBoxesCount = 0;
     }
 
-    if (n <= 0)
+    if (n == 0)
+        return 1;
+
+    if (n < 0)
         return 0;
 
-    ssize_t size = sizeof(SRMRect)*n;
-    connector->damageRects = malloc(size);
-    memcpy(connector->damageRects, rects, size);
-    connector->damageRectsCount = n;
+    ssize_t size = sizeof(SRMBox)*n;
+    connector->damageBoxes = malloc(size);
+
+    for (Int32 i = 0; i < n; i++)
+    {
+        connector->damageBoxes[i].x1 = rects[i].x;
+        connector->damageBoxes[i].y1 = rects[i].y;
+        connector->damageBoxes[i].x2 = rects[i].x + rects[i].width;
+        connector->damageBoxes[i].y2 = rects[i].y + rects[i].height;
+    }
+    connector->damageBoxesCount = n;
     return 1;
 }
+
+UInt8 srmConnectorSetBufferDamageBoxes(SRMConnector *connector, SRMBox *boxes, Int32 n)
+{
+    if (!connector->currentPrimaryPlane || !srmConnectorHasBufferDamageSupport(connector))
+        return 0;
+
+    if (connector->damageBoxes)
+    {
+        free(connector->damageBoxes);
+        connector->damageBoxes = NULL;
+        connector->damageBoxesCount = 0;
+    }
+
+    if (n == 0)
+        return 1;
+
+    if (n < 0)
+        return 0;
+
+    ssize_t size = sizeof(SRMBox)*n;
+    connector->damageBoxes = malloc(size);
+    memcpy(connector->damageBoxes, boxes, size);
+    connector->damageBoxesCount = n;
+    return 1;
+}
+
+SRM_CONNECTOR_SUBPIXEL srmConnectorGetSubPixel(SRMConnector *connector)
+{
+    return connector->subpixel;
+}
+
+UInt64 srmConnectorGetGammaSize(SRMConnector *connector)
+{
+    if (connector->currentCrtc)
+        return srmCrtcGetGammaSize(connector->currentCrtc);
+
+    return 0;
+}
+
+UInt8 srmConnectorSetGamma(SRMConnector *connector, UInt16 *table)
+{
+    if (!connector->currentCrtc)
+    {
+        SRMError("Failed to set gamma for connector %d. Gamma cannot be set on an uninitialized connector.",
+            connector->id);
+        return 0;
+    }
+
+    UInt64 gammaSize = srmCrtcGetGammaSize(connector->currentCrtc);
+
+    if (gammaSize == 0)
+    {
+        SRMError("Failed to set gamma for connector %d. Gamma size is 0, indicating that the driver does not support gamma correction.",
+            connector->id);
+        return 0;
+    }
+
+    if (connector->device->clientCapAtomic && connector->currentCrtc->propIDs.GAMMA_LUT_SIZE)
+    {
+        pthread_mutex_lock(&connector->propsMutex);
+        UInt16 *R = table;
+        UInt16 *G = table + gammaSize;
+        UInt16 *B = G + gammaSize;
+
+        for (UInt64 i = 0; i < gammaSize; i++)
+        {
+            connector->gamma[i].red = R[i];
+            connector->gamma[i].green = G[i];
+            connector->gamma[i].blue = B[i];
+        }
+        
+        connector->atomicChanges |= SRM_ATOMIC_CHANGE_GAMMA_LUT;
+        pthread_mutex_unlock(&connector->propsMutex);
+        pthread_cond_signal(&connector->repaintCond);
+        return 1;
+    }
+    else
+    {
+        if (drmModeCrtcSetGamma(connector->device->fd,
+            connector->currentCrtc->id,
+            (UInt32)gammaSize,
+            table,
+            table + gammaSize,
+            table + gammaSize + gammaSize))
+        {
+            SRMError("Failed to set gamma for connector %d using legacy API drmModeCrtcSetGamma().",
+            connector->id);
+            return 0;
+        }
+    }
+
+    return 1;
+}
+
+UInt8 srmConnectorHasVSyncControlSupport(SRMConnector *connector)
+{
+    return (connector->device->capAsyncPageFlip && !connector->device->clientCapAtomic) ||
+           (connector->device->capAtomicAsyncPageFlip && connector->device->clientCapAtomic);
+}
+
+UInt8 srmConnectorIsVSyncEnabled(SRMConnector *connector)
+{
+    return connector->pendingVSync;
+}
+
+UInt8 srmConnectorEnableVSync(SRMConnector *connector, UInt8 enabled)
+{
+    if (!enabled && !srmConnectorHasVSyncControlSupport(connector))
+        return 0;
+
+    connector->pendingVSync = enabled;
+    return 1;
+}
+
+void srmConnectorSetRefreshRateLimit(SRMConnector *connector, Int32 hz)
+{
+    connector->maxRefreshRate = hz;
+}
+
+Int32 srmConnectorGetRefreshRateLimit(SRMConnector *connector)
+{
+    return connector->maxRefreshRate;
+}
+
+clockid_t srmConnectorGetPresentationClock(SRMConnector *connector)
+{
+    return connector->device->clock;
+}
+
+const SRMPresentationTime *srmConnectorGetPresentationTime(SRMConnector *connector)
+{
+    return &connector->presentationTime;
+}
