@@ -1,7 +1,9 @@
 #ifndef SRMCONNECTOR_H
 #define SRMCONNECTOR_H
 
-#include "SRMTypes.h"
+#include <SRMTypes.h>
+#include <xf86drmMode.h>
+#include <time.h>
 
 #ifdef __cplusplus
 extern "C" {
@@ -564,6 +566,163 @@ UInt8 srmConnectorHasBufferDamageSupport(SRMConnector *connector);
  */
 UInt8 srmConnectorSetBufferDamage(SRMConnector *connector, SRMRect *rects, Int32 n);
 
+/**
+ * @brief Notifies the connector of new damage generated during the last `paintGL()` call using boxes.
+ *
+ * This method is analogous to srmConnectorSetBufferDamage(), but instead of @ref SRMRect, it accepts @ref SRMBox.
+ *
+ * @see srmConnectorSetBufferDamage().
+ *
+ * @param connector Pointer to the @ref SRMConnector structure to notify of buffer damage.
+ * @param boxes An array of @ref SRMBox structures representing the damaged areas.
+ * @param n The number of boxes in the array. Passing 0 unsets the current damage.
+ * @return Returns 1 if the damage notification is successful, 0 if an error occurs.
+ */
+UInt8 srmConnectorSetBufferDamageBoxes(SRMConnector *connector, SRMBox *boxes, Int32 n);
+
+/**
+ * @brief Get the subpixel layout associated with a connector.
+ *
+ * This function retrieves the subpixel layout associated with a given connector.
+ * Subpixels are individual color elements that make up a pixel on a display, and the subpixel layout
+ * is crucial for accurate color interpretation and display. The returned value indicates how the
+ * red, green, and blue subpixels are arranged in relation to each other.
+ *
+ * @note The returned value is only meaningful when a display is connected. If the subpixel layout is unknown or not applicable,
+ *       the method may return @ref SRM_CONNECTOR_SUBPIXEL_UNKNOWN or @ref SRM_CONNECTOR_SUBPIXEL_NONE, respectively.
+ *
+ * @param connector Pointer to the @ref SRMConnector for which the subpixel layout is requested.
+ * @return The subpixel layout associated with the connector.
+ *
+ */
+SRM_CONNECTOR_SUBPIXEL srmConnectorGetSubPixel(SRMConnector *connector);
+
+/**
+ * @brief Gets the number of elements used to represent each RGB gamma correction curve.
+ *
+ * This function retrieves the number of elements (N) used to represent each RGB gamma correction curve,
+ * where N is the count of @ref UInt16 elements for red, green, and blue curves.
+ * 
+ * @note This method should only be called once the connector is assigned a CRTC,
+ *       meaning after the connector is initialized. If called when uninitialized, it returns 0.
+ *       If the connector is initialized and this method returns 0, it indicates that the driver
+ *       does not support gamma correction.
+ * 
+ * @param connector Pointer to the @ref SRMConnector.
+ * @return The number of elements used to represent each RGB gamma correction curve, or 0 if the connector is uninitialized or
+ *         if the driver does not support gamma correction.
+ */
+UInt64 srmConnectorGetGammaSize(SRMConnector *connector);
+
+/**
+ * @brief Sets the gamma correction curves for each RGB component.
+ *
+ * This method allows you to individually set the gamma correction curves for each RGB component.
+ * The number of elements for each curve (N) should be obtained using srmConnectorGetGammaSize().
+ * The table array should then have a size of `3 * N * sizeof(UInt16)` bytes, with N @ref UInt16 values for red,
+ * N for green, and N for blue, in that order. Each value of the curves can represent the full range of @ref UInt16.
+ * 
+ * @note This method must only be called after the connector is initialized.
+ *       If the connector is uninitialized or the driver does not support gamma correction,
+ *       the function returns 0.
+ * 
+ * @param connector Pointer to the @ref SRMConnector.
+ * @param table Pointer to the array containing RGB curves for gamma correction.
+ * @return On success, returns 1. On failure (uninitialized connector or no gamma correction support),
+ *         returns 0. 
+ */
+UInt8 srmConnectorSetGamma(SRMConnector *connector, UInt16 *table);
+
+/**
+ * @brief Checks if the driver supports the ability to turn off vsync.
+ *
+ * If the return value is 0, it indicates that vsync is always enabled.
+ *
+ * @note Disabling vsync for the atomic API is supported only starting from Linux version 6.8.
+ *       If you wish to enforce the use of the legacy API, set the **SRM_FORCE_LEGACY_API** environment variable to 1.
+ *
+ * @param connector The @ref SRMConnector instance.
+ * @return 1 if vsync control is supported, 0 otherwise.
+ */
+UInt8 srmConnectorHasVSyncControlSupport(SRMConnector *connector);
+
+/**
+ * @brief Returns the current vsync status.
+ *
+ * Returns 1 if vsync is enabled, 0 otherwise. V-Sync is enabled by default.
+ *
+ * @param connector The @ref SRMConnector instance.
+ * @return 1 if vsync is enabled, 0 if not.
+ */
+UInt8 srmConnectorIsVSyncEnabled(SRMConnector *connector);
+
+/**
+ * @brief Enable or disable vsync.
+ *
+ * Disabling vsync is only allowed if srmConnectorHasVSyncControlSupport() returns 1.
+ * VSync is enabled by default.
+ *
+ * @note Disabling vsync for the atomic API is supported only starting from Linux version 6.8.
+ *       If you wish to enforce the use of the legacy API, set the **SRM_FORCE_LEGACY_API** environment variable to 1.
+ *
+ * @param connector The @ref SRMConnector instance.
+ * @param enabled Set to 1 to enable vsync, 0 to disable vsync.
+ * @return 1 if the vsync change was successful, 0 otherwise.
+ */
+UInt8 srmConnectorEnableVSync(SRMConnector *connector, UInt8 enabled);
+
+/**
+ * @brief Sets the refresh rate limit when vsync is disabled.
+ *
+ * This function allows controlling the refresh rate limit when vsync is disabled
+ *
+ * @param connector A pointer to the @ref SRMConnector instance for which the refresh rate limit is to be set.
+ * @param hz The desired refresh rate limit in hertz.
+ *           If hz is less than 0, the refresh rate limit is disabled.
+ *           If hz is 0, the maximum refresh rate will be approximately twice the current display mode refresh rate.
+ *           The default value is 0.
+ *
+ * @note Using a value of 0 (twice the current mode refresh rate) provides a good balance between avoiding unnecessary undisplayed frames
+ *       and achieving snappier response times.
+ *
+ * @see srmConnectorGetRefreshRateLimit() to retrieve the current refresh rate limit.
+ *
+ * @warning Disabling the limit may result in visual artifacts and could potentially impact performance.
+ */
+void srmConnectorSetRefreshRateLimit(SRMConnector *connector, Int32 hz);
+
+/**
+ * @brief Retrieves the current refresh rate limit when vsync is disabled.
+ *
+ * This function allows you to query the refresh rate limit that has been previously set
+ * using the srmConnectorSetRefreshRateLimit() function.
+ *
+ * @param connector A pointer to the @ref SRMConnector instance for which the refresh rate limit is to be retrieved.
+ *
+ * @return The current refresh rate limit in hertz.
+ *         If the refresh rate limit is disabled, the returned value will be less than 0.
+ *         If the returned value is 0, means the limit is about twice the current display mode refresh rate.
+ */
+Int32 srmConnectorGetRefreshRateLimit(SRMConnector *connector);
+
+/**
+ * @brief Gets the clock ID used for the timestamps returned by srmConnectorGetPresentationTime().
+ *
+ * The clock ID can be either CLOCK_MONOTONIC or CLOCK_REALTIME.
+ *
+ * @param connector A pointer to the @ref SRMConnector instance.
+ * @return The clock ID used for timestamps.
+ */
+clockid_t srmConnectorGetPresentationClock(SRMConnector *connector);
+
+/**
+ * @brief Retrieves information about how and when the current framebuffer displayed on the screen was presented.
+ *
+ * @param connector A pointer to the @ref SRMConnector instance.
+ * @return Pointer to the structure containing presentation time information, see @ref SRMPresentationTime.
+ */
+const SRMPresentationTime *srmConnectorGetPresentationTime(SRMConnector *connector);
+
 /**
  * @}
  */
