@@ -1,4 +1,4 @@
-#include "private/modes/SRMRenderModeCPU.h"
+#include <private/modes/SRMRenderModeCPU.h>
 #include <private/SRMConnectorPrivate.h>
 #include <private/SRMConnectorModePrivate.h>
 #include <private/SRMDevicePrivate.h>
@@ -30,7 +30,9 @@ SRMConnector *srmConnectorCreate(SRMDevice *device, UInt32 id)
     connector->id = id;
     connector->device = device;
     connector->state = SRM_CONNECTOR_STATE_UNINITIALIZED;
+    connector->pendingVSync = 1;
     pthread_mutex_init(&connector->stateMutex, NULL);
+    pthread_mutex_init(&connector->propsMutex, NULL);
     srmConnectorUpdateProperties(connector);
 
     // srmConnectorUpdateNames(connector) is called after its device is added to the core devices list
@@ -48,6 +50,7 @@ void srmConnectorDestroy(SRMConnector *connector)
     srmConnectorDestroyEncoders(connector);
     srmConnectorDestroyModes(connector);
 
+    pthread_mutex_destroy(&connector->propsMutex);
     pthread_mutex_destroy(&connector->stateMutex);
 
     if (connector->deviceLink)
@@ -94,6 +97,7 @@ UInt8 srmConnectorUpdateProperties(SRMConnector *connector)
         return 0;
     }
 
+    connector->subpixel = (SRM_CONNECTOR_SUBPIXEL)connectorRes->subpixel;
     connector->mmHeight = connectorRes->mmHeight;
     connector->mmWidth = connectorRes->mmWidth;
     connector->connected = connectorRes->connection == DRM_MODE_CONNECTED;
@@ -439,16 +443,20 @@ void *srmConnectorRenderThread(void *conn)
         SRMError("Could not create render mutex for device %s connector %d.",
                  connector->device->name,
                  connector->id);
-        goto repaintMutexFail;
+        goto finish;
     }
+    else
+        connector->hasRepaintMutex = 1;
 
     if (pthread_cond_init(&connector->repaintCond, NULL))
     {
         SRMError("Could not create render cond for device %s connector %d.",
                  connector->device->name,
                  connector->id);
-        goto repaintCondFail;
+        goto finish;
     }
+    else
+        connector->hasRepaintCond = 1;
 
     connector->pendingPageFlip = 0;
 
@@ -460,23 +468,42 @@ void *srmConnectorRenderThread(void *conn)
 
     srmRenderModeCommonCreateCursor(connector);
 
-    if (srmDeviceGetRenderMode(connector->device) == SRM_RENDER_MODE_ITSELF)
+    SRM_RENDER_MODE renderMode = srmDeviceGetRenderMode(connector->device);
+
+    SRMDebug("Connector %d device %s render mode = %s.",
+             connector->id,
+             connector->device->name,
+             srmGetRenderModeString(renderMode));
+
+    if (renderMode == SRM_RENDER_MODE_ITSELF)
         srmRenderModeItselfSetInterface(connector);
-    else if (srmDeviceGetRenderMode(connector->device) == SRM_RENDER_MODE_DUMB)
+    else if (renderMode == SRM_RENDER_MODE_DUMB)
         srmRenderModeDumbSetInterface(connector);
-    else if (srmDeviceGetRenderMode(connector->device) == SRM_RENDER_MODE_CPU)
+    else if (renderMode == SRM_RENDER_MODE_CPU)
         srmRenderModeCPUSetInterface(connector);
     else
-        goto fail;
+    {
+        SRMError("Invalid render mode %s connector %d.",
+                 connector->device->name,
+                 connector->id);
+        goto finish;
+    }
 
     if (!connector->renderInterface.initialize(connector))
-        goto fail;
+    {
+        SRMError("Render mode interface initialize() failed %s connector %d.",
+                 connector->device->name,
+                 connector->id);
+        goto finish;
+    }
 
     connector->renderInitResult = 1;
 
     // Render loop
     while (1)
     {
+        connector->currentVSync = connector->pendingVSync;
+
         if (!srmRenderModeCommonWaitRepaintRequest(connector))
             break;
 
@@ -492,7 +519,7 @@ void *srmConnectorRenderThread(void *conn)
                     pthread_mutex_unlock(&connector->stateMutex);
                     continue;
             }
-            else if (connector->atomicCursorHasChanges)
+            else if (connector->atomicChanges)
             {
                 srmRenderModeCommonPageFlip(connector, connector->lastFb);
             }
@@ -539,12 +566,7 @@ void *srmConnectorRenderThread(void *conn)
         pthread_mutex_unlock(&connector->stateMutex);
     }
 
-fail:
-    srmRenderModeCommonDestroyCursor(connector);
-repaintCondFail:
-    pthread_cond_destroy(&connector->repaintCond);
-repaintMutexFail:
-    pthread_mutex_destroy(&connector->repaintMutex);
+    finish:
     connector->renderInitResult = -1;
     return NULL;
 }
@@ -584,3 +606,75 @@ void srmConnectorSetCursorPlaneToNeededConnector(SRMPlane *cursorPlane)
         }
     }
 }
+
+void srmConnectorRenderThreadCleanUp(SRMConnector *connector)
+{
+    connector->renderThread = 0;
+
+    srmRenderModeCommonDestroyCursor(connector);
+
+    if (connector->hasRepaintCond)
+    {
+        pthread_cond_destroy(&connector->repaintCond);
+        connector->hasRepaintCond = 0;
+    }
+
+    if (connector->hasRepaintMutex)
+    {
+        pthread_mutex_destroy(&connector->repaintMutex);
+        connector->hasRepaintCond = 0;
+    }
+
+    if (connector->currentCrtc)
+    {
+        connector->currentCrtc->currentConnector = NULL;
+        connector->currentCrtc = NULL;
+    }
+
+    if (connector->currentEncoder)
+    {
+        connector->currentEncoder->currentConnector = NULL;
+        connector->currentEncoder = NULL;
+    }
+
+    if (connector->currentPrimaryPlane)
+    {
+        connector->currentPrimaryPlane->currentConnector = NULL;
+        connector->currentPrimaryPlane = NULL;
+    }
+
+    if (connector->currentCursorPlane)
+    {
+        connector->currentCursorPlane->currentConnector = NULL;
+        srmConnectorSetCursorPlaneToNeededConnector(connector->currentCursorPlane);
+        connector->currentCursorPlane = NULL;
+    }
+
+    connector->interfaceData = NULL;
+    connector->interface = NULL;
+
+    if (connector->gamma)
+    {
+        free(connector->gamma);
+        connector->gamma = NULL;
+    }
+
+    if (connector->gammaBlobId)
+    {
+        drmModeDestroyPropertyBlob(connector->device->fd, connector->gammaBlobId);
+        connector->gammaBlobId = 0;
+    }
+
+    if (connector->damageBoxes)
+    {
+        free(connector->damageBoxes);
+        connector->damageBoxes = NULL;
+        connector->damageBoxesCount = 0;
+    }
+
+    if (connector->currentModeBlobId)
+    {
+        drmModeDestroyPropertyBlob(connector->device->fd, connector->currentModeBlobId);
+        connector->currentModeBlobId = 0;
+    }
+}
