@@ -112,7 +112,7 @@ UInt8 srmCoreEnumerateDevices(SRMCore *core)
 
     udev_enumerate_unref(enumerate);
 
-    return 1;
+    return !srmListIsEmpty(core->devices);
 }
 
 UInt8 srmCoreInitMonitor(SRMCore *core)
@@ -210,6 +210,8 @@ SRMDevice *srmCoreFindBestAllocatorDevice(SRMCore *core)
     SRMDevice *bestAllocatorDev = NULL;
     int bestScore = 0;
 
+    const char *preferredDevice = getenv("SRM_ALLOCATOR_DEVICE");
+
     SRMListForeach(item1, core->devices)
     {
         SRMDevice *allocDev = srmListItemGetData(item1);
@@ -217,46 +219,32 @@ SRMDevice *srmCoreFindBestAllocatorDevice(SRMCore *core)
         if (!srmDeviceIsEnabled(allocDev))
             continue;
 
-        if (allocDev->driver == SRM_DEVICE_DRIVER_nouveau && srmListGetLength(core->devices) > 1)
+        int currentScore = 10;
+
+        if (preferredDevice)
+            if (strcmp(preferredDevice, allocDev->name) == 0)
+                return allocDev;
+
+        if (allocDev->driver == SRM_DEVICE_DRIVER_i915)
+            currentScore += 1000;
+        else if (allocDev->driver == SRM_DEVICE_DRIVER_nouveau)
+            currentScore -= 5;
+        else if (allocDev->driver == SRM_DEVICE_DRIVER_nvidia)
+            currentScore -= 5;
+
+        if (!core->forceGlesCPUBufferAllocation)
         {
-            allocDev->capPrimeExport = 0;
-            allocDev->capPrimeImport = 0;
-        }
+            if (srmFormatIsInList(allocDev->dmaRenderFormats, DRM_FORMAT_ARGB8888, DRM_FORMAT_MOD_LINEAR))
+                currentScore += 50;
 
-        int currentScore = srmDeviceGetCapPrimeExport(allocDev) ? 100 : 10;
+            if (srmFormatIsInList(allocDev->dmaRenderFormats, DRM_FORMAT_XRGB8888, DRM_FORMAT_MOD_LINEAR))
+                currentScore += 50;
 
-        SRMListForeach(item2, core->devices)
-        {
-            SRMDevice *otherDev = srmListItemGetData(item2);
+            if (srmFormatIsInList(allocDev->dmaRenderFormats, DRM_FORMAT_BGRA8888, DRM_FORMAT_MOD_LINEAR))
+                currentScore += 25;
 
-            if (!srmDeviceIsEnabled(otherDev) || allocDev == otherDev)
-                continue;
-
-            // GPU can render
-            if (srmDeviceGetCapPrimeExport(allocDev) &&
-                    srmDeviceGetCapPrimeImport(otherDev) &&
-                    dmaFormatsHaveInCommon(allocDev->dmaTextureFormats, otherDev->dmaTextureFormats))
-            {
-
-                currentScore += 100;
-
-                if (allocDev->driver == SRM_DEVICE_DRIVER_nouveau)
-                    currentScore -= 20;
-
-                continue;
-            }
-            // GPU can not render but glRead > dumbBuffer
-            else if (srmDeviceGetCapDumbBuffer(otherDev))
-            {
-                currentScore += 20;
-                continue;
-            }
-            // GPU can not render glRead > CPU > glTex2D > render
-            else
-            {
-                currentScore += 10;
-                continue;
-            }
+            if (srmFormatIsInList(allocDev->dmaRenderFormats, DRM_FORMAT_BGRX8888, DRM_FORMAT_MOD_LINEAR))
+                currentScore += 25;
         }
 
         if (currentScore > bestScore)
@@ -271,64 +259,10 @@ SRMDevice *srmCoreFindBestAllocatorDevice(SRMCore *core)
 
 void srmCoreAssignRendererDevices(SRMCore *core)
 {
-    int enabledDevicesN = 0;
-
-    // Count enabled GPUs
     SRMListForeach(item, core->devices)
     {
         SRMDevice *dev = srmListItemGetData(item);
-        enabledDevicesN += dev->enabled;
-    }
-
-    SRMListForeach(item, core->devices)
-    {
-        SRMDevice *dev = srmListItemGetData(item);
-
-        if (!dev->enabled)
-        {
-            dev->rendererDevice = NULL;
-            continue;
-        }
-
-        // If only one GPU or can import from allocator
-        if (enabledDevicesN == 1 || (dev->capPrimeImport && core->allocatorDevice->capPrimeExport))
-        {
-            dev->rendererDevice = dev;
-            continue;
-        }
-
-        int bestScore = 0;
-        SRMDevice *bestRendererForDev = NULL;
-
-        SRMListForeach(item, core->devices)
-        {
-            SRMDevice *rendererDevice = srmListItemGetData(item);
-
-            if (!rendererDevice->enabled)
-                continue;
-
-            int currentScore = 0;
-
-            if (dev == rendererDevice)
-                currentScore += 10;
-
-            if (!rendererDevice->capDumbBuffer)
-                currentScore += 20;
-
-            if (rendererDevice->capPrimeImport && core->allocatorDevice->capPrimeExport)
-                currentScore += 100;
-
-            if (rendererDevice == core->allocatorDevice)
-                currentScore += 50;
-
-            if (currentScore > bestScore)
-            {
-                bestRendererForDev = rendererDevice;
-                bestScore = currentScore;
-            }
-        }
-
-        dev->rendererDevice = bestRendererForDev;
+        dev->rendererDevice = core->allocatorDevice;
     }
 }
 
@@ -347,17 +281,9 @@ UInt8 srmCoreUpdateBestConfiguration(SRMCore *core)
                    EGL_NO_SURFACE,
                    bestAllocatorDevice->eglSharedContext);
 
-    /* This should be invoked upon GPU hotplug events, but it may affect user resources.
-    if (allocatorDevice && allocatorDevice != bestAllocatorDevice)
-    {
-        // TODO may require update
-    }
-    */
-
     core->allocatorDevice = bestAllocatorDevice;
     srmCoreAssignRendererDevices(core);
     srmCoreUpdateSharedDMATextureFormats(core);
-
     return 1;
 }
 
@@ -551,7 +477,7 @@ static void *srmCoreDeallocatorLoop(void *data)
 
                 if (message->textureID)
                 {
-                    SRMDebug("[%s] GL Texture (%d) deleted.", message->device->name, message->textureID);
+                    SRMDebug("[%s] GL Texture (%d) destroyed.", message->device->name, message->textureID);
                     glDeleteTextures(1, &message->textureID);
                 }
 
@@ -619,7 +545,7 @@ UInt8 srmCoreInitDeallocator(SRMCore *core)
 
     if (pthread_create(&core->deallocatorThread, NULL, srmCoreDeallocatorLoop, core))
     {
-        SRMFatal("[core] Could not start render thread for device %s connector %d.");
+        SRMFatal("[core] Could not start deallocator thread.");
         pthread_mutex_destroy(&core->deallocatorMutex);
         pthread_cond_destroy(&core->deallocatorCond);
         srmListDestroy(core->deallocatorMessages);
@@ -673,170 +599,3 @@ void srmCoreSendDeallocatorMessage(SRMCore *core,
     pthread_cond_signal(&core->deallocatorCond);
     pthread_mutex_unlock(&core->deallocatorMutex);
 }
-
-#if SRM_PAR_CPY == 1
-static void *srmCoreCopyThreadLoop(void *data)
-{
-    struct SRMCopyThread *thread = data;
-    cpu_set_t cpuset;
-    CPU_ZERO(&cpuset);
-    CPU_SET(thread->cpu, &cpuset);
-
-    if (pthread_setaffinity_np(thread->thread, sizeof(cpu_set_t), &cpuset) != 0)
-        SRMDebug("[core] Failed to set copy thread to CPU %d.", thread->cpu);
-    else
-        SRMDebug("[core] New copy thread assigned to CPU %d.", thread->cpu);
-
-    thread->state = 1;
-
-    while (1)
-    {
-        pthread_mutex_lock(&thread->mutex);
-
-        if (thread->finished)
-            pthread_cond_wait(&thread->cond, &thread->mutex);
-
-        if (thread->state == 1)
-        {
-            Int32 a,b;
-            // If not copied yet
-            for (Int32 i = 0; i < thread->rows; i++)
-            {
-                memmove(&thread->dst[ (thread->offsetY + i) * thread->dstStride],
-                       &thread->src[ (thread->offsetY + i) * thread->srcStride],
-                       thread->size);
-            }
-
-            thread->finished = 1;
-        }
-        else
-        {
-            thread->state = 3;
-            pthread_mutex_unlock(&thread->mutex);
-            return NULL;
-        }
-
-        pthread_mutex_unlock(&thread->mutex);
-    }
-}
-
-void srmCoreInitCopyThreads(SRMCore *core)
-{
-    core->copyThreadsCount = sysconf(_SC_NPROCESSORS_ONLN);
-
-    if (core->copyThreadsCount <= 1)
-    {
-        core->copyThreadsCount = 0;
-        return;
-    }
-
-    if (core->copyThreadsCount > SRM_MAX_COPY_THREADS)
-        core->copyThreadsCount = SRM_MAX_COPY_THREADS;
-
-    SRMDebug("[core] CPU cores avaliable: %d.", core->copyThreadsCount);
-
-    pthread_attr_t attr;
-    struct sched_param param;
-
-    // Initialize thread attributes
-    pthread_attr_init(&attr);
-
-    // Set scheduling policy to SCHED_FIFO
-    pthread_attr_setschedpolicy(&attr, SCHED_FIFO);
-
-    // Set the priority using sched_param
-    param.sched_priority = 99;
-    pthread_attr_setschedparam(&attr, &param);
-
-    for (UInt8 i = 0; i < core->copyThreadsCount; i++)
-    {
-        core->copyThreads[i].finished = 1;
-        core->copyThreads[i].core = core;
-        core->copyThreads[i].cpu = i;
-        pthread_mutex_init(&core->copyThreads[i].mutex, NULL);
-        pthread_cond_init(&core->copyThreads[i].cond, NULL);
-        pthread_create(&core->copyThreads[i].thread, &attr, srmCoreCopyThreadLoop, &core->copyThreads[i]);
-    }
-
-    checkThreadsOK:
-
-    for (UInt8 i = 0; i < core->copyThreadsCount; i++)
-    {
-        if (!core->copyThreads[i].state)
-            goto checkThreadsOK;
-    }
-}
-
-// Buffers pointers must already contain (x,y) offset
-UInt8 srmCoreCopy(SRMCore *core, UInt8 *src, UInt8 *dst, Int32 srcStride, Int32 dstStride, Int32 size, Int32 height)
-{
-    if (!core->copyThreadsCount)
-        return 0;
-
-    UInt8 workingThreads = 0;
-
-    struct SRMCopyThread *thread;
-
-    // If there the number of threads is equal or greater the rows to copy
-    // a single row is assigned to each thread
-    if (height <= core->copyThreadsCount)
-    {
-        workingThreads = height;
-
-        for (UInt8 i = 0; i < workingThreads; i++)
-        {
-            thread = &core->copyThreads[i];
-            pthread_mutex_lock(&thread->mutex);
-            thread->finished = 0;
-            thread->src = src;
-            thread->dst = dst;
-            thread->srcStride = srcStride;
-            thread->dstStride = dstStride;
-            thread->size = size;
-            thread->offsetY = i;
-            thread->rows = 1;
-            pthread_cond_signal(&thread->cond);
-            pthread_mutex_unlock(&thread->mutex);
-        }
-    }
-    else
-    {
-        workingThreads = core->copyThreadsCount;
-
-        Int32 rowsPerThread = height / workingThreads;
-        Int32 offsetY = 0;
-
-        for (UInt8 i = 0; i < workingThreads - 1; i++)
-        {
-            thread = &core->copyThreads[i];
-            pthread_mutex_lock(&thread->mutex);
-            thread->finished = 0;
-            thread->src = src;
-            thread->dst = dst;
-            thread->srcStride = srcStride;
-            thread->dstStride = dstStride;
-            thread->size = size;
-            thread->offsetY = offsetY;
-            thread->rows = rowsPerThread;
-            offsetY += rowsPerThread;
-            pthread_cond_signal(&thread->cond);
-            pthread_mutex_unlock(&thread->mutex);
-        }
-
-        thread = &core->copyThreads[workingThreads - 1];
-        pthread_mutex_lock(&thread->mutex);
-        thread->finished = 0;
-        thread->src = src;
-        thread->dst = dst;
-        thread->srcStride = srcStride;
-        thread->dstStride = dstStride;
-        thread->size = size;
-        thread->offsetY = offsetY;
-        thread->rows = height - offsetY;
-        pthread_cond_signal(&thread->cond);
-        pthread_mutex_unlock(&thread->mutex);
-    }
-
-    return 1;
-}
-#endif
