@@ -19,8 +19,43 @@
 #include <xf86drm.h>
 #include <xf86drmMode.h>
 
+static UInt8 srmDeviceInBlacklist(const char *deviceName)
+{
+    char *blacklist = getenv("SRM_DEVICES_BLACKLIST");
+
+    if (!blacklist)
+        return 0;
+
+    size_t extlen = strlen(deviceName);
+    const char *end = blacklist + strlen(blacklist);
+
+    while (blacklist < end)
+    {
+        if (*blacklist == ':')
+        {
+            blacklist++;
+            continue;
+        }
+
+        size_t n = strcspn(blacklist, ":");
+
+        if (n == extlen && strncmp(deviceName, blacklist, n) == 0)
+            return 1;
+
+        blacklist += n;
+    }
+
+    return 0;
+}
+
 SRMDevice *srmDeviceCreate(SRMCore *core, const char *name)
 {
+    if (srmDeviceInBlacklist(name))
+    {
+        SRMWarning("[device] %s is blacklisted. Ignoring it.", name);
+        return NULL;
+    }
+
     // REF 1
     SRMDevice *device = calloc(1, sizeof(SRMDevice));
     strncpy(device->name, name, sizeof(device->name) - 1);
@@ -46,12 +81,16 @@ SRMDevice *srmDeviceCreate(SRMCore *core, const char *name)
 
     if (version)
     {
+        SRMDebug("[%s] DRM Driver: %s.", device->name, version->name);
+
         if (strcmp(version->name, "i915") == 0)
             device->driver = SRM_DEVICE_DRIVER_i915;
         else if (strcmp(version->name, "nouveau") == 0)
             device->driver = SRM_DEVICE_DRIVER_nouveau;
         else if (strcmp(version->name, "lima") == 0)
             device->driver = SRM_DEVICE_DRIVER_lima;
+        else if (strcmp(version->name, "nvidia-drm") == 0)
+            device->driver = SRM_DEVICE_DRIVER_nvidia;
 
         drmFreeVersion(version);
     }
@@ -264,13 +303,14 @@ UInt8 srmDeviceUpdateEGLExtensions(SRMDevice *device)
         return 0;
     }
 
+    SRMDebug("[%s] EGL Extensions: %s.", device->name, extensions);
+
     device->eglExtensions.KHR_image_base = srmEGLHasExtension(extensions, "EGL_KHR_image_base");
     device->eglExtensions.KHR_image = srmEGLHasExtension(extensions, "EGL_KHR_image");
     device->eglExtensions.EXT_image_dma_buf_import = srmEGLHasExtension(extensions, "EGL_EXT_image_dma_buf_import");
     device->eglExtensions.EXT_image_dma_buf_import_modifiers = srmEGLHasExtension(extensions, "EGL_EXT_image_dma_buf_import_modifiers");
     device->eglExtensions.EXT_create_context_robustness = srmEGLHasExtension(extensions, "EGL_EXT_create_context_robustness");
-
-    /* TODO: Which of the above are mandatory ? */
+    device->eglExtensions.KHR_image_pixmap = srmEGLHasExtension(extensions, "EGL_KHR_image_pixmap");
 
     const char *deviceExtensions = NULL, *driverName = NULL;
 
@@ -294,6 +334,8 @@ UInt8 srmDeviceUpdateEGLExtensions(SRMDevice *device)
             goto skipDeviceQuery;
         }
 
+        SRMDebug("[%s] EGL Device Extensions: %s.", device->name, deviceExtensions);
+
         device->eglExtensions.MESA_device_software = srmEGLHasExtension(deviceExtensions, "EGL_MESA_device_software");
 
 #ifdef EGL_DRIVER_NAME_EXT
@@ -352,13 +394,6 @@ UInt8 srmDeviceUpdateDMAFormats(SRMDevice *device)
     }
 
     EGLint formatsCount = 0;
-
-    EGLint fallbackFormats[] =
-    {
-        DRM_FORMAT_ARGB8888,
-        DRM_FORMAT_XRGB8888,
-    };
-
     EGLint *formats = NULL;
     UInt8 allExternalOnly = 1;
 
@@ -366,8 +401,8 @@ UInt8 srmDeviceUpdateDMAFormats(SRMDevice *device)
     {
         if (!device->eglFunctions.eglQueryDmaBufFormatsEXT(device->eglDisplay, 0, NULL, &formatsCount))
         {
-            SRMError("[%s] Failed to query the number of EGL DMA formats, using fallback formats.", device->name);
-            goto fallback;
+            SRMError("[%s] Failed to query the number of EGL DMA formats.", device->name);
+            return 0;
         }
 
         if (formatsCount <= 0)
@@ -387,13 +422,10 @@ UInt8 srmDeviceUpdateDMAFormats(SRMDevice *device)
     }
     else
     {
-        fallback:
-        formatsCount = sizeof(fallbackFormats)/sizeof(fallbackFormats[0]);
-        formats = malloc(sizeof(fallbackFormats));
-        memcpy(formats, fallbackFormats, sizeof(fallbackFormats));
+        SRMError("[%s] Failed to query EGL DMA formats.", device->name);
+        return 0;
     }
 
-
     for (Int32 i = 0; i < formatsCount; i++)
     {
         EGLint modifiersCount = 0;
@@ -435,7 +467,6 @@ UInt8 srmDeviceUpdateDMAFormats(SRMDevice *device)
 
         skipModifiers:
 
-
         for (Int32 j = 0; j < modifiersCount; j++)
         {
             srmFormatsListAddFormat(device->dmaTextureFormats, formats[i], modifiers[j]);
@@ -529,8 +560,11 @@ void srmDeviceUninitializeEGLSharedContext(SRMDevice *device)
 UInt8 srmDeviceUpdateGLExtensions(SRMDevice *device)
 {
     const char *exts = (const char*)glGetString(GL_EXTENSIONS);
+    SRMDebug("[%s] OpenGL Extensions: %s.", device->name, exts);
     device->glExtensions.EXT_read_format_bgra = srmEGLHasExtension(exts, "GL_EXT_read_format_bgra");
     device->glExtensions.EXT_texture_format_BGRA8888 = srmEGLHasExtension(exts, "GL_EXT_texture_format_BGRA8888");
+    device->glExtensions.OES_EGL_image_external = srmEGLHasExtension(exts, "GL_OES_EGL_image_external");
+    device->glExtensions.OES_EGL_image = srmEGLHasExtension(exts, "GL_OES_EGL_image");
     return 1;
 }
 
@@ -553,45 +587,34 @@ UInt8 srmDeviceUpdateClientCaps(SRMDevice *device)
 
 UInt8 srmDeviceUpdateCaps(SRMDevice *device)
 {
-    UInt64 value;
+    UInt64 value = 0;
     drmGetCap(device->fd, DRM_CAP_DUMB_BUFFER, &value);
     device->capDumbBuffer = value == 1;
 
+    value = 0;
     drmGetCap(device->fd, DRM_CAP_PRIME, &value);
-
     device->capPrimeImport = value & DRM_PRIME_CAP_IMPORT;
     device->capPrimeExport = value & DRM_PRIME_CAP_EXPORT;
 
-    // Validate EXPORT cap
-    if (device->capPrimeExport)
-    {
-        bool primeExport = false;
-
-        struct gbm_bo *bo = gbm_bo_create(device->gbm, 256, 256, GBM_FORMAT_ARGB8888, GBM_BO_USE_LINEAR);
-
-        if (bo)
-        {
-            int dma = srmBufferGetDMAFDFromBO(device, bo);
-
-            if (dma != -1)
-            {
-                primeExport = true;
-                close(dma);
-            }
-
-            gbm_bo_destroy(bo);
-        }
-
-        device->capPrimeExport = primeExport;
-    }
-
-
+    value = 0;
     drmGetCap(device->fd, DRM_CAP_ADDFB2_MODIFIERS, &value);
     device->capAddFb2Modifiers = value == 1;
 
+    value = 0;
+    drmGetCap(device->fd, DRM_CAP_TIMESTAMP_MONOTONIC, &value);
+    device->capTimestampMonotonic = value == 1;
+    device->clock = device->capTimestampMonotonic ? CLOCK_MONOTONIC : CLOCK_REALTIME;
+
+    value = 0;
     drmGetCap(device->fd, DRM_CAP_ASYNC_PAGE_FLIP, &value);
     device->capAsyncPageFlip = value == 1;
 
+#ifdef DRM_CAP_ATOMIC_ASYNC_PAGE_FLIP
+    value = 0;
+    drmGetCap(device->fd, DRM_CAP_ATOMIC_ASYNC_PAGE_FLIP, &value);
+    device->capAtomicAsyncPageFlip = value == 1;
+#endif
+
     return 1;
 }
 
@@ -812,7 +835,6 @@ UInt8 srmDeviceHandleHotpluggingEvent(SRMDevice *device)
 
                 // Uninitialize after notify so users can for example backup some data
                 srmConnectorUninitialize(connector);
-
                 srmConnectorUpdateProperties(connector);
                 srmConnectorUpdateNames(connector);
                 srmConnectorUpdateEncoders(connector);
