@@ -9,7 +9,7 @@
 #include <SRMLog.h>
 #include <stdlib.h>
 #include <xf86drmMode.h>
-#include <sys/poll.h>
+#include <poll.h>
 #include <unistd.h>
 
 // Choose EGL configurations
@@ -77,22 +77,76 @@ out:
     return 1;
 }
 
-void srmRenderModeCommonPageFlipHandler(int a, unsigned int b, unsigned int c, unsigned int d, void *data)
+void srmRenderModeCommonPageFlipHandler(Int32 fd, UInt32 seq, UInt32 sec, UInt32 usec, void *data)
 {
-    SRM_UNUSED(a);
-    SRM_UNUSED(b);
-    SRM_UNUSED(c);
-    SRM_UNUSED(d);
+    SRM_UNUSED(fd);
 
     if (data)
     {
         SRMConnector *connector = data;
         connector->pendingPageFlip = 0;
+
+        if (connector->state == SRM_CONNECTOR_STATE_UNINITIALIZED)
+            return;
+
+        if (connector->currentVSync)
+        {
+            connector->presentationTime.flags = SRM_PRESENTATION_TIME_FLAGS_HW_CLOCK |
+                                                SRM_PRESENTATION_TIME_FLAGS_HW_COMPLETION |
+                                                SRM_PRESENTATION_TIME_FLAGS_VSYNC;
+
+            connector->presentationTime.frame = seq;
+            connector->presentationTime.time.tv_sec = sec;
+            connector->presentationTime.time.tv_nsec = usec * 1000;
+            connector->presentationTime.period = connector->currentMode->info.vrefresh == 0 ? 0 : 1000000000/connector->currentMode->info.vrefresh;
+        }
+        else
+        {
+            connector->presentationTime.flags = 0;
+            connector->presentationTime.frame = 0;
+            connector->presentationTime.period = 0;
+
+            Int64 prevUsec = (connector->presentationTime.time.tv_sec * 1000000LL) + (connector->presentationTime.time.tv_nsec / 1000LL);
+            clock_gettime(connector->device->clock, &connector->presentationTime.time);
+
+            if (connector->maxRefreshRate < 0)
+                return;
+
+            Int64 currUsec = (connector->presentationTime.time.tv_sec * 1000000LL) + (connector->presentationTime.time.tv_nsec / 1000LL);
+
+            Int64 periodUsec;
+
+            // Limit FPS to 2 * vrefresh
+            if (connector->maxRefreshRate == 0)
+                periodUsec = (connector->currentMode->info.vrefresh == 0 ? 0 : 490000/(connector->currentMode->info.vrefresh));
+            else
+                periodUsec = (connector->currentMode->info.vrefresh == 0 ? 0 : 1000000/(connector->maxRefreshRate));
+
+            if (periodUsec > 0)
+            {
+                Int64 diffUsec = currUsec - prevUsec;
+
+                if (diffUsec >= 0 && diffUsec < periodUsec)
+                {
+                    usleep(periodUsec - diffUsec);
+                    clock_gettime(connector->device->clock, &connector->presentationTime.time);
+                }
+            }
+
+        }
     }
 }
 
 UInt8 srmRenderModeCommonCreateCursor(SRMConnector *connector)
 {
+    if (connector->device->driver == SRM_DEVICE_DRIVER_nvidia)
+    {
+        char *env = getenv("SRM_NVIDIA_CURSOR");
+
+        if (!env || atoi(env) != 1)
+            return 0;
+    }
+
     connector->cursorBO = gbm_bo_create(connector->device->gbm,
                                         64,
                                         64,
@@ -187,9 +241,9 @@ UInt8 srmRenderModeCommonWaitRepaintRequest(SRMConnector *connector)
 {
     pthread_mutex_lock(&connector->repaintMutex);
 
-    if ((!connector->repaintRequested && !connector->atomicCursorHasChanges) || srmCoreIsSuspended(connector->device->core))
+    if ((!connector->repaintRequested && !connector->atomicChanges) || srmCoreIsSuspended(connector->device->core))
     {
-        connector->atomicCursorHasChanges = 0;
+        connector->atomicChanges = 0;
         pthread_cond_wait(&connector->repaintCond, &connector->repaintMutex);
     }
 
@@ -199,11 +253,12 @@ UInt8 srmRenderModeCommonWaitRepaintRequest(SRMConnector *connector)
     if (connector->state == SRM_CONNECTOR_STATE_UNINITIALIZING)
     {
         pthread_mutex_unlock(&connector->stateMutex);
+        connector->pendingPageFlip = 1;
+        srmRenderModeCommonWaitPageFlip(connector, 3);
         connector->interface->uninitializeGL(connector, connector->interfaceData);
         connector->renderInterface.uninitialize(connector);
-        pthread_mutex_lock(&connector->stateMutex);
+        eglReleaseThread();
         connector->state = SRM_CONNECTOR_STATE_UNINITIALIZED;
-        pthread_mutex_unlock(&connector->stateMutex);
         return 0;
     }
     pthread_mutex_unlock(&connector->stateMutex);
@@ -211,108 +266,147 @@ UInt8 srmRenderModeCommonWaitRepaintRequest(SRMConnector *connector)
     return 1;
 }
 
-void srmRenderModeCommitCursorChanges(SRMConnector *connector, drmModeAtomicReqPtr req)
+void srmRenderModeCommitAtomicChanges(SRMConnector *connector, drmModeAtomicReqPtr req, UInt8 clearFlags)
 {
-    if (!connector->currentCursorPlane)
-        return;
+    pthread_mutex_lock(&connector->propsMutex);
 
-    UInt8 updatedFB = 0;
-
-    if (connector->atomicCursorHasChanges & SRM_CURSOR_ATOMIC_CHANGE_BUFFER)
+    if (connector->currentCursorPlane)
     {
-        struct gbm_bo *tmpBo = connector->cursorBO;
-        connector->cursorBO = connector->cursorBOPending;
-        connector->cursorBOPending = tmpBo;
+        UInt8 updatedFB = 0;
 
-        UInt32 tmpFb = connector->cursorFB;
-        connector->cursorFB = connector->cursorFBPending;
-        connector->cursorFBPending = tmpFb;
-
-        if (connector->cursorVisible)
+        if (connector->atomicChanges & SRM_ATOMIC_CHANGE_CURSOR_BUFFER)
         {
+            if (clearFlags)
+                connector->atomicChanges &= ~SRM_ATOMIC_CHANGE_CURSOR_BUFFER;
+
+            struct gbm_bo *tmpBo = connector->cursorBO;
+            connector->cursorBO = connector->cursorBOPending;
+            connector->cursorBOPending = tmpBo;
+
+            UInt32 tmpFb = connector->cursorFB;
+            connector->cursorFB = connector->cursorFBPending;
+            connector->cursorFBPending = tmpFb;
+
+            if (connector->cursorVisible)
+            {
+                drmModeAtomicAddProperty(req,
+                                        connector->currentCursorPlane->id,
+                                        connector->currentCursorPlane->propIDs.FB_ID,
+                                        connector->cursorFB);
+                updatedFB = 1;
+            }
+        }
+
+        if (connector->atomicChanges & SRM_ATOMIC_CHANGE_CURSOR_VISIBILITY)
+        {
+            if (clearFlags)
+                connector->atomicChanges &= ~SRM_ATOMIC_CHANGE_CURSOR_VISIBILITY;
+
+            if (connector->cursorVisible)
+            {
+                if (!updatedFB)
+                {
+                    drmModeAtomicAddProperty(req,
+                                            connector->currentCursorPlane->id,
+                                            connector->currentCursorPlane->propIDs.FB_ID,
+                                            connector->cursorFB);
+                }
+
+                drmModeAtomicAddProperty(req,
+                                        connector->currentCursorPlane->id,
+                                        connector->currentCursorPlane->propIDs.CRTC_ID,
+                                        connector->currentCrtc->id);
+
+                drmModeAtomicAddProperty(req,
+                                        connector->currentCursorPlane->id,
+                                        connector->currentCursorPlane->propIDs.CRTC_W,
+                                        64);
+
+                drmModeAtomicAddProperty(req,
+                                        connector->currentCursorPlane->id,
+                                        connector->currentCursorPlane->propIDs.CRTC_H,
+                                        64);
+
+                drmModeAtomicAddProperty(req,
+                                        connector->currentCursorPlane->id,
+                                        connector->currentCursorPlane->propIDs.SRC_X,
+                                        0);
+
+                drmModeAtomicAddProperty(req,
+                                        connector->currentCursorPlane->id,
+                                        connector->currentCursorPlane->propIDs.SRC_Y,
+                                        0);
+
+                drmModeAtomicAddProperty(req,
+                                        connector->currentCursorPlane->id,
+                                        connector->currentCursorPlane->propIDs.SRC_W,
+                                        (UInt64)64 << 16);
+
+                drmModeAtomicAddProperty(req,
+                                        connector->currentCursorPlane->id,
+                                        connector->currentCursorPlane->propIDs.SRC_H,
+                                        (UInt64)64 << 16);
+            }
+            else
+            {
+                drmModeAtomicAddProperty(req,
+                                        connector->currentCursorPlane->id,
+                                        connector->currentCursorPlane->propIDs.CRTC_ID,
+                                        0);
+
+                drmModeAtomicAddProperty(req,
+                                        connector->currentCursorPlane->id,
+                                        connector->currentCursorPlane->propIDs.FB_ID,
+                                        0);
+            }
+        }
+
+        if (connector->atomicChanges & SRM_ATOMIC_CHANGE_CURSOR_POSITION)
+        {
+            if (clearFlags)
+                connector->atomicChanges &= ~SRM_ATOMIC_CHANGE_CURSOR_POSITION;
+
             drmModeAtomicAddProperty(req,
-                                     connector->currentCursorPlane->id,
-                                     connector->currentCursorPlane->propIDs.FB_ID,
-                                     connector->cursorFB);
-            updatedFB = 1;
+                                    connector->currentCursorPlane->id,
+                                    connector->currentCursorPlane->propIDs.CRTC_X,
+                                    connector->cursorX);
+
+            drmModeAtomicAddProperty(req,
+                                    connector->currentCursorPlane->id,
+                                    connector->currentCursorPlane->propIDs.CRTC_Y,
+                                    connector->cursorY);
         }
     }
 
-    if (connector->atomicCursorHasChanges & SRM_CURSOR_ATOMIC_CHANGE_VISIBILITY)
+    if (connector->atomicChanges & SRM_ATOMIC_CHANGE_GAMMA_LUT)
     {
-        if (connector->cursorVisible)
+        if (clearFlags)
+            connector->atomicChanges &= ~SRM_ATOMIC_CHANGE_GAMMA_LUT;
+
+        if (connector->gammaBlobId)
         {
-            if (!updatedFB)
-            {
-                drmModeAtomicAddProperty(req,
-                                         connector->currentCursorPlane->id,
-                                         connector->currentCursorPlane->propIDs.FB_ID,
-                                         connector->cursorFB);
-            }
-
-            drmModeAtomicAddProperty(req,
-                                     connector->currentCursorPlane->id,
-                                     connector->currentCursorPlane->propIDs.CRTC_ID,
-                                     connector->currentCrtc->id);
-
-            drmModeAtomicAddProperty(req,
-                                     connector->currentCursorPlane->id,
-                                     connector->currentCursorPlane->propIDs.CRTC_W,
-                                     64);
-
-            drmModeAtomicAddProperty(req,
-                                     connector->currentCursorPlane->id,
-                                     connector->currentCursorPlane->propIDs.CRTC_H,
-                                     64);
-
-            drmModeAtomicAddProperty(req,
-                                     connector->currentCursorPlane->id,
-                                     connector->currentCursorPlane->propIDs.SRC_X,
-                                     0);
-
-            drmModeAtomicAddProperty(req,
-                                     connector->currentCursorPlane->id,
-                                     connector->currentCursorPlane->propIDs.SRC_Y,
-                                     0);
-
-            drmModeAtomicAddProperty(req,
-                                     connector->currentCursorPlane->id,
-                                     connector->currentCursorPlane->propIDs.SRC_W,
-                                     (UInt64)64 << 16);
-
-            drmModeAtomicAddProperty(req,
-                                     connector->currentCursorPlane->id,
-                                     connector->currentCursorPlane->propIDs.SRC_H,
-                                     (UInt64)64 << 16);
+            drmModeDestroyPropertyBlob(connector->device->fd, connector->gammaBlobId);
+            connector->gammaBlobId = 0;
         }
+
+        if (drmModeCreatePropertyBlob(connector->device->fd, 
+            connector->gamma,
+            srmCrtcGetGammaSize(connector->currentCrtc) * sizeof(*connector->gamma),
+            &connector->gammaBlobId))
+        {
+            connector->gammaBlobId = 0;
+            SRMError("Failed to create gamma lut blob for connector %d.", connector->id);
+	    }
         else
         {
             drmModeAtomicAddProperty(req,
-                                     connector->currentCursorPlane->id,
-                                     connector->currentCursorPlane->propIDs.CRTC_ID,
-                                     0);
-
-            drmModeAtomicAddProperty(req,
-                                     connector->currentCursorPlane->id,
-                                     connector->currentCursorPlane->propIDs.FB_ID,
-                                     0);
+                connector->currentCrtc->id,
+                connector->currentCrtc->propIDs.GAMMA_LUT,
+                connector->gammaBlobId);
         }
     }
 
-    if (connector->atomicCursorHasChanges & SRM_CURSOR_ATOMIC_CHANGE_POSITION)
-    {
-        drmModeAtomicAddProperty(req,
-                                 connector->currentCursorPlane->id,
-                                 connector->currentCursorPlane->propIDs.CRTC_X,
-                                 connector->cursorX);
-
-        drmModeAtomicAddProperty(req,
-                                 connector->currentCursorPlane->id,
-                                 connector->currentCursorPlane->propIDs.CRTC_Y,
-                                 connector->cursorY);
-    }
-
-    connector->atomicCursorHasChanges = 0;
+    pthread_mutex_unlock(&connector->propsMutex);
 }
 
 void srmRenderModeCommonDestroyCursor(SRMConnector *connector)
@@ -363,16 +457,19 @@ void srmRenderModeCommonDestroyCursor(SRMConnector *connector)
         connector->cursorBOPending = NULL;
     }
 
-    connector->atomicCursorHasChanges = 0;
+    connector->atomicChanges &= ~(SRM_ATOMIC_CHANGE_CURSOR_VISIBILITY | SRM_ATOMIC_CHANGE_CURSOR_POSITION | SRM_ATOMIC_CHANGE_CURSOR_BUFFER);
     connector->cursorVisible = 0;
 }
 
-Int32 srmRenderModeAtomicCommit(Int32 fd, drmModeAtomicReqPtr req, UInt32 flags, void *data)
+Int32 srmRenderModeAtomicCommit(Int32 fd, drmModeAtomicReqPtr req, UInt32 flags, void *data, UInt8 forceRetry)
 {
+    if (!forceRetry)
+        return drmModeAtomicCommit(fd, req, flags, data);
+
     Int32 ret;
 
     retry:
-    ret = drmModeAtomicCommit(fd, req, flags, data);
+    ret = drmModeAtomicCommit(fd, req, flags | DRM_MODE_ATOMIC_TEST_ONLY, data);
 
     // -EBUSY
     if (ret == -16)
@@ -381,7 +478,7 @@ Int32 srmRenderModeAtomicCommit(Int32 fd, drmModeAtomicReqPtr req, UInt32 flags,
         goto retry;
     }
 
-    return ret;
+    return drmModeAtomicCommit(fd, req, flags, data);
 }
 
 Int32 srmRenderModeCommonUpdateMode(SRMConnector *connector, UInt32 fb)
@@ -409,17 +506,21 @@ Int32 srmRenderModeCommonUpdateMode(SRMConnector *connector, UInt32 fb)
             // Set mode
             drmModeAtomicReqPtr req = drmModeAtomicAlloc();
 
-            UInt32 modeID;
+            if (connector->currentModeBlobId)
+            {
+                drmModeDestroyPropertyBlob(connector->device->fd, connector->currentModeBlobId);
+                connector->currentModeBlobId = 0;
+            }
 
             drmModeCreatePropertyBlob(connector->device->fd,
                                       &connector->currentMode->info,
                                       sizeof(drmModeModeInfo),
-                                      &modeID);
+                                      &connector->currentModeBlobId);
 
             drmModeAtomicAddProperty(req,
                                      connector->currentCrtc->id,
                                      connector->currentCrtc->propIDs.MODE_ID,
-                                     modeID);
+                                     connector->currentModeBlobId);
 
             drmModeAtomicAddProperty(req,
                                      connector->currentCrtc->id,
@@ -490,12 +591,12 @@ Int32 srmRenderModeCommonUpdateMode(SRMConnector *connector, UInt32 fb)
                                      connector->currentPrimaryPlane->propIDs.SRC_H,
                                      (UInt64)connector->currentMode->info.vdisplay << 16);
 
-            srmRenderModeCommitCursorChanges(connector, req);
+            srmRenderModeCommitAtomicChanges(connector, req, 1);
 
             ret = srmRenderModeAtomicCommit(connector->device->fd,
                                 req,
                                 DRM_MODE_ATOMIC_ALLOW_MODESET,
-                                connector);
+                                connector, 1);
 
             if (ret)
             {
@@ -504,12 +605,14 @@ Int32 srmRenderModeCommonUpdateMode(SRMConnector *connector, UInt32 fb)
                          connector->id, ret);
             }
 
-            drmModeDestroyPropertyBlob(connector->device->fd, modeID);
-
             drmModeAtomicFree(req);
         }
         else
         {
+            connector->pendingPageFlip = 1;
+            srmRenderModeCommonWaitPageFlip(connector, 3);
+            connector->pendingPageFlip = 0;
+
             int ret;
 
             drmModeSetCrtc(connector->device->fd,
@@ -560,6 +663,10 @@ Int32 srmRenderModeCommonUpdateMode(SRMConnector *connector, UInt32 fb)
         }
         else
         {
+            connector->pendingPageFlip = 1;
+            srmRenderModeCommonWaitPageFlip(connector, 3);
+            connector->pendingPageFlip = 0;
+
             ret = drmModeSetCrtc(connector->device->fd,
                                connector->currentCrtc->id,
                                0,
@@ -662,17 +769,21 @@ void srmRenderModeCommonResumeRendering(SRMConnector *connector, UInt32 fb)
         drmModeAtomicReqPtr req;
         req = drmModeAtomicAlloc();
 
-        UInt32 modeID;
+        if (connector->currentModeBlobId)
+        {
+            drmModeDestroyPropertyBlob(connector->device->fd, connector->currentModeBlobId);
+            connector->currentModeBlobId = 0;
+        }
 
         drmModeCreatePropertyBlob(connector->device->fd,
                                   &connector->currentMode->info,
                                   sizeof(drmModeModeInfo),
-                                  &modeID);
+                                  &connector->currentModeBlobId);
 
         drmModeAtomicAddProperty(req,
                                  connector->currentCrtc->id,
                                  connector->currentCrtc->propIDs.MODE_ID,
-                                 modeID);
+                                 connector->currentModeBlobId);
 
         drmModeAtomicAddProperty(req,
                                  connector->currentCrtc->id,
@@ -743,13 +854,13 @@ void srmRenderModeCommonResumeRendering(SRMConnector *connector, UInt32 fb)
                                  connector->currentPrimaryPlane->propIDs.SRC_H,
                                  (UInt64)connector->currentMode->info.vdisplay << 16);
 
-        srmRenderModeCommitCursorChanges(connector, req);
+        srmRenderModeCommitAtomicChanges(connector, req, 1);
 
         // Commit
         ret = srmRenderModeAtomicCommit(connector->device->fd,
                                         req,
                                         DRM_MODE_ATOMIC_ALLOW_MODESET,
-                                        connector);
+                                        connector, 1);
 
         drmModeAtomicFree(req);
 
@@ -785,15 +896,114 @@ Int32 srmRenderModeCommonInitCrtc(SRMConnector *connector, UInt32 fb)
     Int32 ret;
     connector->lastFb = fb;
 
-    if (connector->state == SRM_CONNECTOR_STATE_INITIALIZING)
+    if (connector->device->clientCapAtomic)
     {
-        connector->interface->initializeGL(connector,
-                                           connector->interfaceData);
-    }
-    else if (connector->state == SRM_CONNECTOR_STATE_CHANGING_MODE)
-    {
-        connector->interface->resizeGL(connector,
-                                       connector->interfaceData);
+        drmModeAtomicReqPtr req;
+        req = drmModeAtomicAlloc();
+
+        if (connector->currentModeBlobId)
+        {
+            drmModeDestroyPropertyBlob(connector->device->fd, connector->currentModeBlobId);
+            connector->currentModeBlobId = 0;
+        }
+
+        drmModeCreatePropertyBlob(connector->device->fd,
+                                  &connector->currentMode->info,
+                                  sizeof(drmModeModeInfo),
+                                  &connector->currentModeBlobId);
+
+        drmModeAtomicAddProperty(req,
+                                 connector->currentCrtc->id,
+                                 connector->currentCrtc->propIDs.MODE_ID,
+                                 connector->currentModeBlobId);
+
+        drmModeAtomicAddProperty(req,
+                                 connector->currentCrtc->id,
+                                 connector->currentCrtc->propIDs.ACTIVE,
+                                 1);
+
+        // Connector
+
+        drmModeAtomicAddProperty(req,
+                                 connector->id,
+                                 connector->propIDs.CRTC_ID,
+                                 connector->currentCrtc->id);
+
+        drmModeAtomicAddProperty(req,
+                                 connector->id,
+                                 connector->propIDs.link_status,
+                                 DRM_MODE_LINK_STATUS_GOOD);
+
+        // Plane
+
+        drmModeAtomicAddProperty(req,
+                                 connector->currentPrimaryPlane->id,
+                                 connector->currentPrimaryPlane->propIDs.CRTC_ID,
+                                 connector->currentCrtc->id);
+
+        drmModeAtomicAddProperty(req,
+                                 connector->currentPrimaryPlane->id,
+                                 connector->currentPrimaryPlane->propIDs.CRTC_X,
+                                 0);
+
+        drmModeAtomicAddProperty(req,
+                                 connector->currentPrimaryPlane->id,
+                                 connector->currentPrimaryPlane->propIDs.CRTC_Y,
+                                 0);
+
+        drmModeAtomicAddProperty(req,
+                                 connector->currentPrimaryPlane->id,
+                                 connector->currentPrimaryPlane->propIDs.CRTC_W,
+                                 connector->currentMode->info.hdisplay);
+
+        drmModeAtomicAddProperty(req,
+                                 connector->currentPrimaryPlane->id,
+                                 connector->currentPrimaryPlane->propIDs.CRTC_H,
+                                 connector->currentMode->info.vdisplay);
+
+        drmModeAtomicAddProperty(req,
+                                 connector->currentPrimaryPlane->id,
+                                 connector->currentPrimaryPlane->propIDs.FB_ID,
+                                 fb);
+
+        drmModeAtomicAddProperty(req,
+                                 connector->currentPrimaryPlane->id,
+                                 connector->currentPrimaryPlane->propIDs.SRC_X,
+                                 0);
+
+        drmModeAtomicAddProperty(req,
+                                 connector->currentPrimaryPlane->id,
+                                 connector->currentPrimaryPlane->propIDs.SRC_Y,
+                                 0);
+
+        drmModeAtomicAddProperty(req,
+                                 connector->currentPrimaryPlane->id,
+                                 connector->currentPrimaryPlane->propIDs.SRC_W,
+                                 (UInt64)connector->currentMode->info.hdisplay << 16);
+
+        drmModeAtomicAddProperty(req,
+                                 connector->currentPrimaryPlane->id,
+                                 connector->currentPrimaryPlane->propIDs.SRC_H,
+                                 (UInt64)connector->currentMode->info.vdisplay << 16);
+
+        srmRenderModeCommitAtomicChanges(connector, req, 1);
+
+        // Commit
+        ret = srmRenderModeAtomicCommit(connector->device->fd,
+                                        req,
+                                        DRM_MODE_ATOMIC_ALLOW_MODESET,
+                                        connector, 1);
+
+        drmModeAtomicFree(req);
+
+        if (ret)
+        {
+            SRMError("Failed to set crtc mode on device %s connector %d (atomic).",
+                    connector->device->name,
+                    connector->id);
+        }
+        else
+            goto skipLegacy;
     }
 
     /* Occasionally, the Atomic API fails to set the connector CRTC for reasons unknown.
@@ -815,79 +1025,129 @@ Int32 srmRenderModeCommonInitCrtc(SRMConnector *connector, UInt32 fb)
                     connector->id);
         return 0;
     }
+
+    skipLegacy:
+
+    if (connector->state == SRM_CONNECTOR_STATE_INITIALIZING)
+    {
+        connector->interface->initializeGL(connector,
+                                           connector->interfaceData);
+        glFinish();
+    }
+    else if (connector->state == SRM_CONNECTOR_STATE_CHANGING_MODE)
+    {
+        connector->interface->resizeGL(connector,
+                                       connector->interfaceData);
+        glFinish();
+    }
     
     return 1;
 }
 
 void srmRenderModeCommonPageFlip(SRMConnector *connector, UInt32 fb)
 {
-    Int32 ret;
+    Int32 ret = 0;
 
     UInt32 buffersCount = srmConnectorGetBuffersCount(connector);
 
-    if (buffersCount == 1 || buffersCount > 2)
-    {
-        struct pollfd fds;
-        fds.fd = connector->device->fd;
-        fds.events = POLLIN;
-        fds.revents = 0;
-
-        while (connector->pendingPageFlip)
-        {
-            if (connector->state != SRM_CONNECTOR_STATE_INITIALIZED)
-                break;
-
-            // Prevent multiple threads invoking the drmHandleEvent at a time wich causes bugs
-            // If more than 1 connector is requesting a page flip, both can be handled here
-            // since the struct passed to drmHandleEvent is standard and could be handling events
-            // from any connector (E.g. pageFlipHandler(conn1) or pageFlipHandler(conn2))
-            pthread_mutex_lock(&connector->device->pageFlipMutex);
-
-            // Double check if the pageflip was notified in another thread
-            if (!connector->pendingPageFlip)
-            {
-                pthread_mutex_unlock(&connector->device->pageFlipMutex);
-                break;
-            }
-
-            poll(&fds, 1, 500);
-
-            if(fds.revents & POLLIN)
-                drmHandleEvent(fds.fd, &connector->drmEventCtx);
-
-            pthread_mutex_unlock(&connector->device->pageFlipMutex);
-        }
-    }
+    if (connector->pendingPageFlip || buffersCount == 1 || buffersCount > 2)
+        srmRenderModeCommonWaitPageFlip(connector, -1);
 
     connector->lastFb = fb;
-    connector->pendingPageFlip = 1;
 
     if (connector->device->clientCapAtomic)
     {
-        drmModeAtomicReqPtr req;
-        req = drmModeAtomicAlloc();
-        srmRenderModeCommitCursorChanges(connector, req);
+        if (connector->currentVSync)
+        {
+            drmModeAtomicReqPtr req;
+            req = drmModeAtomicAlloc();
+            srmRenderModeCommitAtomicChanges(connector, req, 0);
+            drmModeAtomicAddProperty(req,
+                                     connector->currentPrimaryPlane->id,
+                                     connector->currentPrimaryPlane->propIDs.FB_ID,
+                                     connector->lastFb);
+            ret = srmRenderModeAtomicCommit(connector->device->fd,
+                                        req,
+                                        DRM_MODE_PAGE_FLIP_EVENT | DRM_MODE_ATOMIC_NONBLOCK,
+                                        connector, 1);
+            if (ret == 0 && !connector->firstPageFlip)
+                connector->atomicChanges = 0;
 
-        drmModeAtomicAddProperty(req,
-                                 connector->currentPrimaryPlane->id,
-                                 connector->currentPrimaryPlane->propIDs.FB_ID,
-                                 connector->lastFb);
+            drmModeAtomicFree(req);
+            connector->pendingPageFlip = 1;
+        }
+        else
+        {
+            if (connector->atomicChanges == 0)
+            {
+                drmModeAtomicReqPtr req;
+                req = drmModeAtomicAlloc();
+                drmModeAtomicAddProperty(req,
+                                         connector->currentPrimaryPlane->id,
+                                         connector->currentPrimaryPlane->propIDs.FB_ID,
+                                         connector->lastFb);
+                ret = srmRenderModeAtomicCommit(connector->device->fd,
+                                                req,
+                                                DRM_MODE_PAGE_FLIP_EVENT | DRM_MODE_PAGE_FLIP_ASYNC | DRM_MODE_ATOMIC_NONBLOCK,
+                                                connector, 1);
+                drmModeAtomicFree(req);
+            }
 
-        ret = srmRenderModeAtomicCommit(connector->device->fd,
-                                  req,
-                                  DRM_MODE_PAGE_FLIP_EVENT | DRM_MODE_ATOMIC_NONBLOCK,
-                                  connector);
-        drmModeAtomicFree(req);
+            if (connector->atomicChanges || ret == -22)
+            {
+                drmModeAtomicReqPtr req;
+                req = drmModeAtomicAlloc();
+                srmRenderModeCommitAtomicChanges(connector, req, 0);
+                drmModeAtomicAddProperty(req,
+                                         connector->currentPrimaryPlane->id,
+                                         connector->currentPrimaryPlane->propIDs.FB_ID,
+                                         connector->lastFb);
+                ret = srmRenderModeAtomicCommit(connector->device->fd,
+                                                req,
+                                                DRM_MODE_PAGE_FLIP_EVENT | DRM_MODE_ATOMIC_NONBLOCK,
+                                                connector, 1);
+                drmModeAtomicFree(req);
+
+                // Clear flags on success
+                if (ret == 0 && !connector->firstPageFlip)
+                    connector->atomicChanges = 0;
+            }
+
+            connector->pendingPageFlip = 1;
+        }
     }
     else
     {
-        ret = drmModePageFlip(connector->device->fd,
-                              connector->currentCrtc->id,
-                              connector->lastFb,
-                              DRM_MODE_PAGE_FLIP_EVENT,
-                              connector);
+        if (connector->currentVSync)
+        {
+            ret = drmModePageFlip(connector->device->fd,
+                                  connector->currentCrtc->id,
+                                  connector->lastFb,
+                                  DRM_MODE_PAGE_FLIP_EVENT,
+                                  connector);
+        }
+        else
+        {
+            ret = drmModePageFlip(connector->device->fd,
+                                  connector->currentCrtc->id,
+                                  connector->lastFb,
+                                  DRM_MODE_PAGE_FLIP_ASYNC | DRM_MODE_PAGE_FLIP_EVENT,
+                                  connector);
+
+            if (ret == -22)
+            {
+                ret = drmModePageFlip(connector->device->fd,
+                                      connector->currentCrtc->id,
+                                      connector->lastFb,
+                                      DRM_MODE_PAGE_FLIP_EVENT,
+                                      connector);
+            }
+        }
+
+        connector->pendingPageFlip = 1;
     }
 
+
     if (ret)
     {
         connector->pendingPageFlip = 0;
@@ -899,58 +1159,61 @@ void srmRenderModeCommonPageFlip(SRMConnector *connector, UInt32 fb)
     if (buffersCount == 2 || connector->firstPageFlip)
     {
         connector->firstPageFlip = 0;
+        srmRenderModeCommonWaitPageFlip(connector, -1);
+    }
+}
 
-        struct pollfd fds;
-        fds.fd = connector->device->fd;
-        fds.events = POLLIN;
-        fds.revents = 0;
+void srmRenderModeCommonWaitPageFlip(SRMConnector *connector, Int32 iterLimit)
+{
+    struct pollfd fds;
+    fds.fd = connector->device->fd;
+    fds.events = POLLIN;
+    fds.revents = 0;
 
-        while(connector->pendingPageFlip)
+    while (connector->pendingPageFlip)
+    {
+        if (connector->state != SRM_CONNECTOR_STATE_INITIALIZED || iterLimit == 0)
+            break;
+
+        // Prevent multiple threads invoking the drmHandleEvent at a time wich causes bugs
+        // If more than 1 connector is requesting a page flip, both can be handled here
+        // since the struct passed to drmHandleEvent is standard and could be handling events
+        // from any connector (E.g. pageFlipHandler(conn1) or pageFlipHandler(conn2))
+        pthread_mutex_lock(&connector->device->pageFlipMutex);
+
+        // Double check if the pageflip was notified in another thread
+        if (!connector->pendingPageFlip)
         {
-            if (connector->state != SRM_CONNECTOR_STATE_INITIALIZED)
-                break;
-
-            // Prevent multiple threads invoking the drmHandleEvent at a time wich causes bugs
-            // If more than 1 connector is requesting a page flip, both can be handled here
-            // since the struct passed to drmHandleEvent is standard and could be handling events
-            // from any connector (E.g. pageFlipHandler(conn1) or pageFlipHandler(conn2))
-            pthread_mutex_lock(&connector->device->pageFlipMutex);
-
-            // Double check if the pageflip was notified in another thread
-            if (!connector->pendingPageFlip)
-            {
-                pthread_mutex_unlock(&connector->device->pageFlipMutex);
-                break;
-            }
-
-            poll(&fds, 1, 500);
-
-            if(fds.revents & POLLIN)
-                drmHandleEvent(fds.fd, &connector->drmEventCtx);
-
             pthread_mutex_unlock(&connector->device->pageFlipMutex);
+            break;
         }
+
+        poll(&fds, 1, iterLimit == -1 ? 500 : 1);
+        drmHandleEvent(fds.fd, &connector->drmEventCtx);
+
+        if (iterLimit > 0)
+            iterLimit--;
+
+        pthread_mutex_unlock(&connector->device->pageFlipMutex);
     }
 }
 
 Int32 srmRenderModeAtomicResetConnectorProps(SRMConnector *connector)
 {
+    connector->pendingPageFlip = 1;
+    srmRenderModeCommonWaitPageFlip(connector, 3);
+    connector->pendingPageFlip = 0;
+
     Int32 ret;
     drmModeAtomicReqPtr req;
     req = drmModeAtomicAlloc();
-    UInt32 modeID;
 
     // Unset mode
 
-    drmModeCreatePropertyBlob(connector->device->fd,
-                              &connector->currentMode->info,
-                              0,
-                              &modeID);
-
     drmModeAtomicAddProperty(req,
                              connector->currentCrtc->id,
                              connector->currentCrtc->propIDs.MODE_ID,
-                             modeID);
+                             0);
 
     drmModeAtomicAddProperty(req,
                              connector->currentCrtc->id,
@@ -1021,14 +1284,12 @@ Int32 srmRenderModeAtomicResetConnectorProps(SRMConnector *connector)
                              connector->currentPrimaryPlane->propIDs.SRC_H,
                              0);
 
-    srmRenderModeCommitCursorChanges(connector, req);
+    srmRenderModeCommitAtomicChanges(connector, req, 1);
 
     ret = srmRenderModeAtomicCommit(connector->device->fd,
                                     req,
                                     DRM_MODE_ATOMIC_ALLOW_MODESET,
-                                    connector);
-
-    drmModeDestroyPropertyBlob(connector->device->fd, modeID);
+                                    connector, 1);
 
     drmModeAtomicFree(req);
 
